import json

import torch
import torch.nn as nn
from torch.nn import functional as F
import transformers
from transformers import BertModel, BertPreTrainedModel


class ModifiedFocalLoss(nn.Module):
    def __init__(self, alpha=0, gamma=0.7, reduction='mean', eps=1e-6):
        super().__init__()
        if reduction not in ['mean', 'none', 'sum']:
            raise NotImplementedError('Reduction {} not implemented.'.format(reduction))
        self.reduction = reduction
        self.alpha = alpha
        self.gamma = gamma
        self.eps = eps

    def forward(self, probs, target):
        probs = torch.clamp(probs, min=self.eps, max=1-self.eps)
        target = F.one_hot(target, num_classes=probs.size(1))
        p_t = torch.where(target == 1, probs, 1 - probs)
        losses = - 1 * (1 + self.alpha - p_t) ** self.gamma * torch.log(p_t)
        return self._reduce(losses)

    def _reduce(self, losses):
        if self.reduction == 'mean':
            return losses.mean()
        elif self.reduction == 'sum':
            return losses.sum()
        else:
            return losses


class G2PW(BertPreTrainedModel):
    def __init__(self, model_source, labels, chars, pos_tags,
                 use_conditional=False, param_conditional=None,
                 use_focal=False, param_focal=None,
                 use_pos=False, param_pos=None):
        super().__init__(model_source)

        self.num_labels = len(labels)
        self.num_chars = len(chars)
        self.num_pos_tags = len(pos_tags)

        self.bert = BertModel(self.config)

        self.classifier = nn.Linear(self.config.hidden_size, self.num_labels)

        self.use_conditional = use_conditional
        self.param_conditional = param_conditional
        if self.use_conditional:
            conditional_affect_location = self.param_conditional['affect_location']
            target_size = self.config.hidden_size if conditional_affect_location == 'emb' else self.num_labels

            if self.param_conditional['bias']:
                self.descriptor_bias = nn.Embedding(1, target_size)
            if self.param_conditional['char-linear']:
                self.char_descriptor = nn.Embedding(self.num_chars, target_size)
            if self.param_conditional['pos-linear']:
                self.pos_descriptor = nn.Embedding(self.num_pos_tags, target_size)
            if self.param_conditional['char+pos-second']:
                self.second_order_descriptor = nn.Embedding(self.num_chars * self.num_pos_tags, target_size)
            if self.param_conditional['char+pos-second_lowrank']:
                assert not self.param_conditional['char+pos-second']
                assert 0 < self.param_conditional['lowrank_size'] < target_size
                self.second_lowrank_descriptor = nn.Sequential(
                    nn.Embedding(self.num_chars * self.num_pos_tags, self.param_conditional['lowrank_size']),
                    nn.Linear(self.param_conditional['lowrank_size'], target_size)
                )
            if self.param_conditional['char+pos-second_fm']:
                assert not self.param_conditional['char+pos-second']
                assert 0 < self.param_conditional['fm_size']
                self.second_fm_char_emb = nn.Sequential(
                    nn.Embedding(self.num_chars, self.param_conditional['fm_size'] * target_size),
                    nn.Unflatten(1, (target_size, self.param_conditional['fm_size']))
                )
                self.second_fm_pos_emb = nn.Sequential(
                    nn.Embedding(self.num_pos_tags, self.param_conditional['fm_size'] * target_size),
                    nn.Unflatten(1, (target_size, self.param_conditional['fm_size']))
                )
            if self.param_conditional['fix_mode']:
                assert all([not self.param_conditional[x] for x in ['bias', 'char-linear', 'pos-linear', 'char+pos-second', 'char+pos-second_lowrank', 'char+pos-second_fm']])
                assert self.param_conditional['affect_location'] == 'softmax'
                count_dict = json.load(open(self.param_conditional['count_json']))
                if self.param_conditional['fix_mode'] == 'count_distr:char':
                    char_fix_count = torch.tensor(
                        [[count_dict['by_char'][char].get(label, 0.) for label in labels] for char in chars]
                    )
                    self.char_fix_emb = nn.parameter.Parameter(
                        char_fix_count / char_fix_count.sum(dim=-1, keepdim=True),
                        requires_grad=False)
                elif self.param_conditional['fix_mode'] == 'count_distr:char+pos':
                    char_pos_fix_count = torch.tensor(
                        [[count_dict['by_char_pos'][f'{char}-{pos}'].get(label, 0.)
                          if f'{char}-{pos}' in count_dict['by_char_pos'] else 0.
                          for label in labels]
                         for char in chars for pos in pos_tags]
                    )
                    self.char_pos_fix_emb = nn.parameter.Parameter(
                        char_pos_fix_count / char_pos_fix_count.sum(dim=-1, keepdim=True),
                        requires_grad=False)
                else:
                    raise Exception

        self.use_focal = use_focal
        self.param_focal = param_focal

        self.use_pos = use_pos
        self.param_pos = param_pos
        if self.use_pos and self.param_pos['pos_joint_training']:
            self.pos_classifier = nn.Linear(self.config.hidden_size, self.num_pos_tags)

    def _weighted_softmax(self, logits, weights, eps):
        max_logits, _ = torch.max(logits, dim=-1, keepdim=True)
        weighted_exp_logits = torch.exp(logits - max_logits) * weights
        norm = torch.sum(weighted_exp_logits, dim=-1, keepdim=True)
        probs = weighted_exp_logits / norm
        probs = torch.clamp(probs, min=eps, max=1-eps)
        return probs

    def _get_char_pos_ids(self, char_ids, pos_ids):
        return char_ids * self.num_pos_tags + pos_ids

    def _get_pos_loss_scaling_when_using_focal(self, phoneme_probs, label_ids):
        phoneme_probs = phoneme_probs.detach()
        phoneme_target = F.one_hot(label_ids, num_classes=phoneme_probs.size(1))
        phoneme_p_t = torch.where(phoneme_target == 1, phoneme_probs, 1 - phoneme_probs)
        avg_phoneme_p_t = phoneme_p_t.mean()
        scaling = (1 + self.param_focal['alpha'] - avg_phoneme_p_t) ** self.param_focal['gamma']
        return scaling

    def forward(self, input_ids, token_type_ids, attention_mask, phoneme_mask, char_ids, position_ids, pos_ids=None, label_ids=None, eps=1e-6):
        transformers_major_ver = int(transformers.__version__.split('.')[0])
        if transformers_major_ver >= 4:
            sequence_output, pooled_output = self.bert(
                input_ids,
                token_type_ids=token_type_ids,
                attention_mask=attention_mask,
                return_dict=False
            )
        else:
            sequence_output, pooled_output = self.bert(
                input_ids,
                token_type_ids=token_type_ids,
                attention_mask=attention_mask
            )

        batch_size = input_ids.size(0)
        orig_selected_hidden = sequence_output[torch.arange(batch_size), position_ids]
        selected_hidden = orig_selected_hidden

        if self.use_conditional:
            if (self.param_conditional['char+pos-second']
                    or self.param_conditional['char+pos-second_lowrank']
                    or self.param_conditional['char+pos-second_fm']
                    or self.param_conditional['pos-linear']
                    or self.param_conditional['fix_mode'] == 'count_distr:char+pos'):
                pred_pos_ids = pos_ids if self.training or not self.param_pos['pos_joint_training'] \
                    else self.pos_classifier(orig_selected_hidden).argmax(dim=-1)  # teacher mode while training

            affect_terms = []
            if self.param_conditional['bias']:
                bias_tensor = self.descriptor_bias(torch.zeros_like(char_ids))
                affect_terms.append(bias_tensor)
            if self.param_conditional['char-linear']:
                affect_terms.append(self.char_descriptor(char_ids))
            if self.param_conditional['pos-linear']:
                affect_terms.append(self.pos_descriptor(pred_pos_ids))
            if self.param_conditional['char+pos-second']:
                char_pos_ids = self._get_char_pos_ids(char_ids, pred_pos_ids)
                affect_terms.append(self.second_order_descriptor(char_pos_ids))
            if self.param_conditional['char+pos-second_lowrank']:
                char_pos_ids = self._get_char_pos_ids(char_ids, pred_pos_ids)
                affect_terms.append(self.second_lowrank_descriptor(char_pos_ids))
            if self.param_conditional['char+pos-second_fm']:
                affect_terms.append(
                    torch.sum(
                        self.second_fm_char_emb(char_ids) * self.second_fm_pos_emb(pred_pos_ids),
                        dim=-1
                    )
                )
            affect_hidden = sum(affect_terms)

            if self.param_conditional['fix_mode'] == 'count_distr:char':
                phoneme_mask = phoneme_mask * F.embedding(char_ids, self.char_fix_emb)
            elif self.param_conditional['fix_mode'] == 'count_distr:char+pos':
                char_pos_ids = self._get_char_pos_ids(char_ids, pred_pos_ids)
                phoneme_mask = phoneme_mask * F.embedding(char_pos_ids, self.char_pos_fix_emb)
            elif self.param_conditional['affect_location'] == 'emb':
                selected_hidden = selected_hidden * affect_hidden
            elif self.param_conditional['affect_location'] == 'softmax':
                phoneme_mask = phoneme_mask * torch.sigmoid(affect_hidden)
            else:
                raise Exception

        logits = self.classifier(selected_hidden)
        probs = self._weighted_softmax(logits, phoneme_mask, eps)
        if label_ids is not None:
            if self.use_focal:
                loss_layer = ModifiedFocalLoss(alpha=self.param_focal['alpha'], gamma=self.param_focal['gamma'])
                loss = loss_layer(probs, label_ids)
            else:
                loss_layer = nn.NLLLoss()
                log_probs = torch.log(probs)
                loss = loss_layer(log_probs, label_ids)

            pos_logits = None
            if self.use_pos and pos_ids is not None and self.param_pos['pos_joint_training']:
                pos_logits = self.pos_classifier(orig_selected_hidden)
                loss_fct = nn.CrossEntropyLoss()
                pos_loss = loss_fct(pos_logits, pos_ids)
                scaling = self._get_pos_loss_scaling_when_using_focal(probs, label_ids) if self.use_focal else 1.
                loss += self.param_pos['weight'] * scaling * pos_loss

            return probs, loss, pos_logits
        else:
            return probs



import torch
from torch.utils.data import Dataset
from torch.nn.utils.rnn import pad_sequence

from g2pw.utils import tokenize_and_map

ANCHOR_CHAR = '▁'


def prepare_data(sent_path, lb_path=None):
    raw_texts = open(sent_path).read().rstrip().split('\n')
    query_ids = [raw.index(ANCHOR_CHAR) for raw in raw_texts]
    texts = [raw.replace(ANCHOR_CHAR, '') for raw in raw_texts]
    if lb_path is None:
        return texts, query_ids
    else:
        phonemes = open(lb_path).read().rstrip().split('\n')
        return texts, query_ids, phonemes


def get_phoneme_labels(polyphonic_chars):
    labels = sorted(list(set([phoneme for char, phoneme in polyphonic_chars])))
    char2phonemes = {}
    for char, phoneme in polyphonic_chars:
        if char not in char2phonemes:
            char2phonemes[char] = []
        char2phonemes[char].append(labels.index(phoneme))
    return labels, char2phonemes


def get_char_phoneme_labels(polyphonic_chars):
    labels = sorted(list(set([f'{char} {phoneme}' for char, phoneme in polyphonic_chars])))
    char2phonemes = {}
    for char, phoneme in polyphonic_chars:
        if char not in char2phonemes:
            char2phonemes[char] = []
        char2phonemes[char].append(labels.index(f'{char} {phoneme}'))
    return labels, char2phonemes


def prepare_pos(pos_path):
     return open(pos_path).read().rstrip().split('\n')


class TextDataset(Dataset):
    POS_TAGS = ['UNK', 'A', 'C', 'D', 'I', 'N', 'P', 'T', 'V', 'DE', 'SHI']

    def __init__(self, tokenizer, labels, char2phonemes, chars, texts, query_ids, phonemes=None, pos_tags=None,
                 use_mask=False, use_char_phoneme=False, use_pos=False, window_size=None, max_len=512, for_train=True):
        self.tokenizer = tokenizer
        self.max_len = max_len
        self.window_size = window_size
        self.for_train = for_train

        self.labels = labels
        self.char2phonemes = char2phonemes
        self.chars = chars
        self.texts = texts
        self.query_ids = query_ids
        self.phonemes = phonemes
        self.pos_tags = pos_tags

        self.use_mask = use_mask
        self.use_char_phoneme = use_char_phoneme
        self.use_pos = use_pos

        if window_size is not None:
            self.truncated_texts, self.truncated_query_ids = self._truncate_texts(self.window_size, texts, query_ids)

    def _truncate_texts(self, window_size, texts, query_ids):
        truncated_texts = []
        truncated_query_ids = []
        for text, query_id in zip(texts, query_ids):
            start = max(0, query_id - window_size // 2)
            end = min(len(text), query_id + window_size // 2)
            truncated_text = text[start:end]
            truncated_texts.append(truncated_text)

            truncated_query_id = query_id - start
            truncated_query_ids.append(truncated_query_id)
        return truncated_texts, truncated_query_ids

    def _truncate(self, max_len, text, query_id, tokens, text2token, token2text):
        truncate_len = max_len - 2
        if len(tokens) <= truncate_len:
            return (text, query_id, tokens, text2token, token2text)

        token_position = text2token[query_id]

        token_start = token_position - truncate_len // 2
        token_end = token_start + truncate_len
        font_exceed_dist = -token_start
        back_exceed_dist = token_end - len(tokens)
        if font_exceed_dist > 0:
            token_start += font_exceed_dist
            token_end += font_exceed_dist
        elif back_exceed_dist > 0:
            token_start -= back_exceed_dist
            token_end -= back_exceed_dist

        start = token2text[token_start][0]
        end = token2text[token_end - 1][1]

        return (
            text[start:end],
            query_id - start,
            tokens[token_start:token_end],
            [i - token_start if i is not None else None for i in text2token[start:end]],
            [(s - start, e - start) for s, e in token2text[token_start:token_end]]
        )

    def __getitem__(self, idx):
        text = (self.truncated_texts if self.window_size else self.texts)[idx].lower()
        query_id = (self.truncated_query_ids if self.window_size else self.query_ids)[idx]

        try:
            tokens, text2token, token2text = tokenize_and_map(self.tokenizer, text)
        except Exception:
            print(f'warning: text "{text}" is invalid')
            return self[(idx + 1) % len(self)]

        text, query_id, tokens, text2token, token2text = self._truncate(self.max_len, text, query_id, tokens, text2token, token2text)

        processed_tokens = ['[CLS]'] + tokens + ['[SEP]']

        input_ids = torch.tensor(self.tokenizer.convert_tokens_to_ids(processed_tokens))
        token_type_ids = torch.tensor([0] * len(processed_tokens))
        attention_mask = torch.tensor([1] * len(processed_tokens))

        query_char = text[query_id]
        phoneme_mask = [1 if i in self.char2phonemes[query_char] else 0 for i in range(len(self.labels))] \
            if self.use_mask else [1] * len(self.labels)
        char_id = self.chars.index(query_char)
        position_id = text2token[query_id] + 1  # [CLS] token locate at first place

        outputs = {
            'input_ids': input_ids,
            'token_type_ids': token_type_ids,
            'attention_mask': attention_mask,
            'phoneme_mask': phoneme_mask,
            'char_id': char_id,
            'position_id': position_id,
        }

        if self.use_pos and self.pos_tags is not None:
            pos_id = self.POS_TAGS.index(self.pos_tags[idx])
            outputs['pos_id'] = pos_id

        if self.for_train:
            phoneme = self.phonemes[idx]
            label_id = self.labels.index(f'{query_char} {phoneme}' if self.use_char_phoneme else phoneme)
            outputs['label_id'] = label_id

        info = {
            'text': text,
            'tokens': tokens,
            'text2token': text2token,
            'token2text': token2text
        }
        outputs['info'] = info
        return outputs

    def __len__(self):
        return len(self.texts)

    def create_mini_batch(self, samples):

        def _agg(name):
            return [sample[name] for sample in samples]

        # zero pad 到同一序列長度
        input_ids = pad_sequence(_agg('input_ids'), batch_first=True)
        token_type_ids = pad_sequence(_agg('token_type_ids'), batch_first=True)
        attention_mask = pad_sequence(_agg('attention_mask'), batch_first=True)
        phoneme_mask = torch.tensor(_agg('phoneme_mask'), dtype=torch.float)
        char_ids = torch.tensor(_agg('char_id'), dtype=torch.long)
        position_ids = torch.tensor(_agg('position_id'), dtype=torch.long)

        batch_output = {
            'input_ids': input_ids,
            'token_type_ids': token_type_ids,
            'attention_mask': attention_mask,
            'phoneme_mask': phoneme_mask,
            'char_ids': char_ids,
            'position_ids': position_ids
        }

        if self.use_pos and self.pos_tags is not None:
            pos_ids = torch.tensor(_agg('pos_id'), dtype=torch.long)
            batch_output['pos_ids'] = pos_ids

        if self.for_train:
            label_ids = torch.tensor(_agg('label_id'), dtype=torch.long)
            batch_output['label_ids'] = label_ids
        else:
            infos = _agg('info')
            batch_output['infos'] = infos

        return batch_output



import os
import json
import requests
import zipfile
from io import BytesIO
import shutil

import torch
from torch.utils.data import DataLoader
from transformers import BertTokenizer
from tqdm import tqdm

try:
    from opencc import OpenCC
except:
    pass

from g2pw.module import G2PW
from g2pw.dataset import prepare_data, TextDataset, get_phoneme_labels, get_char_phoneme_labels
from g2pw.utils import load_config

MODEL_URL = 'https://storage.googleapis.com/esun-ai/g2pW/G2PWModel-v2.zip'


def predict(model, dataloader, device, labels, turnoff_tqdm=False):
    model.eval()

    all_preds = []
    all_confidences = []
    with torch.no_grad():
        generator = dataloader if turnoff_tqdm else tqdm(dataloader, desc='predict')
        for data in generator:
            input_ids, token_type_ids, attention_mask, phoneme_mask, char_ids, position_ids = \
                [data[name].to(device) for name in ('input_ids', 'token_type_ids', 'attention_mask', 'phoneme_mask', 'char_ids', 'position_ids')]

            probs = model(
                input_ids=input_ids,
                token_type_ids=token_type_ids,
                attention_mask=attention_mask,
                phoneme_mask=phoneme_mask,
                char_ids=char_ids,
                position_ids=position_ids
            )

            max_probs, preds = map(lambda x: x.cpu().tolist(), probs.max(dim=-1))
            all_preds += [labels[pred] for pred in preds]
            all_confidences += max_probs

    return all_preds, all_confidences


def download_model(model_dir):
    os.makedirs(model_dir, exist_ok=True)

    r = requests.get(MODEL_URL, allow_redirects=True)
    zip_file = zipfile.ZipFile(BytesIO(r.content))

    for member in zip_file.namelist():
        filename = os.path.basename(member)
        # skip directories
        if not filename:
            continue

        # copy file (taken from zipfile's extract)
        source = zip_file.open(member)
        target = open(os.path.join(model_dir, filename), "wb")
        with source, target:
            shutil.copyfileobj(source, target)


class G2PWConverter:
    def __init__(self, model_dir='G2PWModel/', style='bopomofo', model_source=None, use_cuda=False, num_workers=None, batch_size=None,
                 turnoff_tqdm=True, enable_non_tradional_chinese=False):
        if not os.path.exists(os.path.join(model_dir, 'best_accuracy.pth')):
            download_model(model_dir)

        self.config = load_config(os.path.join(model_dir, 'config.py'), use_default=True)

        self.num_workers = num_workers if num_workers else self.config.num_workers
        self.batch_size = batch_size if batch_size else self.config.batch_size
        self.model_source = model_source if model_source else self.config.model_source
        self.turnoff_tqdm = turnoff_tqdm
        self.enable_opencc = enable_non_tradional_chinese

        self.device = torch.device('cuda' if use_cuda else 'cpu')

        self.tokenizer = BertTokenizer.from_pretrained(self.config.model_source)

        polyphonic_chars_path = os.path.join(model_dir, 'POLYPHONIC_CHARS.txt')
        monophonic_chars_path = os.path.join(model_dir, 'MONOPHONIC_CHARS.txt')
        self.polyphonic_chars = [line.split('\t') for line in open(polyphonic_chars_path).read().strip().split('\n')]
        self.monophonic_chars = [line.split('\t') for line in open(monophonic_chars_path).read().strip().split('\n')]
        self.labels, self.char2phonemes = get_char_phoneme_labels(self.polyphonic_chars) if self.config.use_char_phoneme else get_phoneme_labels(self.polyphonic_chars)

        self.chars = sorted(list(self.char2phonemes.keys()))
        self.pos_tags = TextDataset.POS_TAGS

        self.model = G2PW.from_pretrained(
            self.model_source,
            labels=self.labels,
            chars=self.chars,
            pos_tags=self.pos_tags,
            use_conditional=self.config.use_conditional,
            param_conditional=self.config.param_conditional,
            use_focal=self.config.use_focal,
            param_focal=self.config.param_focal,
            use_pos=self.config.use_pos,
            param_pos=self.config.param_pos
        )
        checkpoint = os.path.join(model_dir, 'best_accuracy.pth')
        self.model.load_state_dict(torch.load(checkpoint, map_location=self.device))
        self.model.to(self.device)

        with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               'bopomofo_to_pinyin_wo_tune_dict.json'), 'r') as fr:
            self.bopomofo_convert_dict = json.load(fr)
        self.style_convert_func = {
            'bopomofo': lambda x: x,
            'pinyin': self._convert_bopomofo_to_pinyin,
        }[style]

        with open(os.path.join(os.path.dirname(os.path.abspath(__file__)),
                               'char_bopomofo_dict.json'), 'r') as fr:
            self.char_bopomofo_dict = json.load(fr)

        if self.enable_opencc:
            self.cc = OpenCC('s2tw')

    def _convert_bopomofo_to_pinyin(self, bopomofo):
        tone = bopomofo[-1]
        assert tone in '12345'
        component = self.bopomofo_convert_dict.get(bopomofo[:-1])
        if component:
            return component + tone
        else:
            print(f'Warning: "{bopomofo}" cannot convert to pinyin')
            return None

    def __call__(self, sentences):
        if isinstance(sentences, str):
            sentences = [sentences]

        if self.enable_opencc:
            translated_sentences = []
            for sent in sentences:
                translated_sent = self.cc.convert(sent)
                assert len(translated_sent) == len(sent)
                translated_sentences.append(translated_sent)
            sentences = translated_sentences

        texts, query_ids, sent_ids, partial_results = self._prepare_data(sentences)
        if len(texts) == 0:
            # sentences no polyphonic words
            return partial_results

        dataset = TextDataset(self.tokenizer, self.labels, self.char2phonemes, self.chars, texts, query_ids,
                              use_mask=self.config.use_mask, use_char_phoneme=self.config.use_char_phoneme,
                              window_size=self.config.window_size, for_train=False)

        dataloader = DataLoader(
            dataset=dataset,
            batch_size=self.batch_size,
            collate_fn=dataset.create_mini_batch,
            num_workers=self.num_workers
        )

        preds, confidences = predict(self.model, dataloader, self.device, self.labels, turnoff_tqdm=self.turnoff_tqdm)
        if self.config.use_char_phoneme:
            preds = [pred.split(' ')[1] for pred in preds]

        results = partial_results
        for sent_id, query_id, pred in zip(sent_ids, query_ids, preds):
            results[sent_id][query_id] = self.style_convert_func(pred)

        return results

    def _prepare_data(self, sentences):
        polyphonic_chars = set(self.chars)
        monophonic_chars_dict = {
            char: phoneme for char, phoneme in self.monophonic_chars
        }
        texts, query_ids, sent_ids, partial_results = [], [], [], []
        for sent_id, sent in enumerate(sentences):
            partial_result = [None] * len(sent)
            for i, char in enumerate(sent):
                if char in polyphonic_chars:
                    texts.append(sent)
                    query_ids.append(i)
                    sent_ids.append(sent_id)
                elif char in monophonic_chars_dict:
                    partial_result[i] =  self.style_convert_func(monophonic_chars_dict[char])
                elif char in self.char_bopomofo_dict:
                    partial_result[i] =  self.style_convert_func(self.char_bopomofo_dict[char][0])
            partial_results.append(partial_result)
        return texts, query_ids, sent_ids, partial_results
